/******************************************************************
   Hra had
   Autor: Zdeněk Šesták 2016

   Klasická hra využívající sériovou linku pro terminálový výstup.
   Výhodné při použití např. s putty.exe
 *****************************************************************/

// ==================== Konfigurace hry ====================
// Parametry hracího pole
const int WIDTH = 60;
const int HEIGHT = 20;

// Maximální délka hada
const int SNAKE_MAX = 100;

// Maximální počet jedů
const int POISON_MAX = 100;

// Rychlost komunikace
const long SERIAL_SPEED = 250000;

// ======================= Proměnné ========================
// skóre
int score;

// x-ové souřadnice těla hada
byte body_x[SNAKE_MAX];
// y-ové souřadnice těla hada
byte body_y[SNAKE_MAX];
// délka těla
int body_len;

// x-ové souřadnice jedu
byte poison_x[POISON_MAX];
// y-ové souřadnice jedu
byte poison_y[POISON_MAX];
// počet jedů
int poison_count;

// směr pohybu
const int MOVE_DOWN   = 0;
const int MOVE_UP     = 1;
const int MOVE_LEFT   = 2;
const int MOVE_RIGHT  = 3;
int moveDir = MOVE_DOWN;

// návratové hodnoty funkce moveSnake
const int ID_NOTHING = 0;
const int ID_WALL   = 1;
const int ID_SNAKE  = 2;
const int ID_LEAF   = 3;
const int ID_POISON = 4;

// souřadnice pokrmu
int leaf_x;
int leaf_y;

// stav čekání na stisknutí klávesy, po kterém následuje restart hry
bool waitState = true;

//-----------------------------------------------------------
// clearScreen - smaže obrazovku
//-----------------------------------------------------------
void clearScreen()
{
  Serial.print("\033[2J");
  // skryje kurzor
  Serial.print(F("\033[?25l"));
}


//-----------------------------------------------------------
// setCursor - nastaví kurzor na pozici x y
//-----------------------------------------------------------
void setCursor(byte x, byte y )
{
  Serial.print("\033[");
  Serial.print(y);
  Serial.write(';');
  Serial.print(x);
  Serial.write('H');
}

//-----------------------------------------------------------
// drawFrame - vykreslení rámu
//-----------------------------------------------------------
void drawFrame()
{
  // smaže obrazovku
  clearScreen();
  // kurzor na začátek obrazovky
  setCursor(1, 1);
  // vykreslí horní čáru
  for (int i = 0; i < WIDTH + 2; i++)
  {
    Serial.print(F("▓"));
  }
  Serial.println();
  // vykreslí prostřední část
  for (int j = 0; j < HEIGHT; j++)
  {
    Serial.print(F("▓"));
    for (int i = 0; i < WIDTH; i++)
    {
      Serial.print(F(" "));
    }
    Serial.println(F("▓"));
  }
  // vykreslí spodní čáru
  for (int i = 0; i < WIDTH + 2; i++)
  {
    Serial.print(F("▓"));
  }

}

//-----------------------------------------------------------
// newLeaf - vygeneruje pozici nového lístku, zkontroluje
// zda na nové pozici není had nebo něco jiného
//-----------------------------------------------------------
void newLeaf()
{
  bool ok = true;
  do
  {
    leaf_x = random(2, WIDTH + 1);
    leaf_y = random(2, HEIGHT + 1);
    // otestuje, zda na této pozici není tělo hada
    ok = true;
    for (int i = 0; i < body_len; i++)
    {
      if (leaf_x == body_x[i] && leaf_y == body_y[i])
      {
        // na vygenerovaných souřadnicích již něco je,
        // vynutíme si tak nové generování 
        ok = false;
        break;
      }
    }
  } while (!ok);
  setCursor(leaf_x, leaf_y);
  Serial.print(F("\033[32m♣\033[0m"));
}

//-----------------------------------------------------------
// newPoison - vygeneruje pozici nového jedu, zkontroluje
// zda na nové pozici není had nebo něco jiného
//-----------------------------------------------------------
void newPoison()
{
  // pokud už je dosaženo maxima jedů, nic dále negenerujeme
  if (poison_count == POISON_MAX)
    return;
  bool ok = true;
  int px, py;
  do
  {
    px = random(2, WIDTH + 1);
    py = random(2, HEIGHT + 1);
    // otestuje, zda na této pozici není tělo hada
    ok = true;
    for (int i = 0; i < body_len; i++)
    {
      if (px == body_x[i] && py == body_y[i])
      {
        ok = false;
        break;
      }
    }
    // otestuje, zda na této pozici není lístek
    if (px == leaf_x && py == leaf_y)
      ok = false;
  } while (!ok);
  // vloží nový jed do seznamu
  poison_x[poison_count] = px;
  poison_y[poison_count] = py;
  poison_count++;
  setCursor(px, py);
  Serial.print(F("\033[31m♦\033[0m"));
}

//-----------------------------------------------------------
// restartGame - inicializuje novou hru
// 1) vykreslí hrací plochu
// 2) nastaví hada na 3 políčka uprostřed plochy + vykreslí jej
// 3) vygeneruje lístek
// 4) nastaví směr nahoru
// 5) nuluje skóre
//-----------------------------------------------------------
void restartGame()
{
  drawFrame();
  body_len = 3;
  body_x[0] = WIDTH / 2;
  body_y[0] = HEIGHT / 2;
  body_x[1] = WIDTH / 2;
  body_y[1] = HEIGHT / 2 + 1;
  body_x[2] = WIDTH / 2;
  body_y[2] = HEIGHT / 2 + 2;
  moveDir = MOVE_UP;
  // vykreslení hada
  // první souřadnice je hlava
  setCursor( body_x[0], body_y[0]);
  Serial.print(F("☺"));
  // zbytek tělo
  for (int i = 1; i < body_len; i++)
  {
    setCursor( body_x[i], body_y[i]);
    Serial.print(F("O"));
  }
  poison_count = 0;
  newLeaf();
  score = 0;
}

//-----------------------------------------------------------
// moveSnake - posune hada o jedno políčko
// Funkce vrátí kód symbolu, do kterého se had zakousl
//-----------------------------------------------------------
int moveSnake()
{
  // odmaže ocas
  setCursor(body_x[body_len - 1], body_y[body_len - 1]);
  Serial.write(' ');

  int hx = body_x[0];
  int hy = body_y[0];
  // vypočte se nová souřadnice hlavy
  switch (moveDir)
  {
    case MOVE_DOWN:
      hy++;
      break;
    case MOVE_UP:
      hy--;
      break;
    case MOVE_LEFT:
      hx--;
      break;
    case MOVE_RIGHT:
      hx++;
      break;
  }
  // určí, do čeho se had zakousne
  int r = ID_NOTHING;
  if ( (hx < 2) || (hx > WIDTH + 1) || (hy < 2) || (hy > HEIGHT + 1))
  {
    r = ID_WALL;
  }
  else if (hx == leaf_x && hy == leaf_y)
  {
    r = ID_LEAF;
  }
  else
  {
    for (int i = 0; i < body_len; i++)
    {
      if (hx == body_x[i] && hy == body_y[i])
      {
        r = ID_SNAKE;
        break;
      }
    }
    for (int i = 0; i < poison_count; i++)
    {
      if (hx == poison_x[i] && hy == poison_y[i])
      {
        r = ID_POISON;
        break;
      }
    }
  }

  // hlava se nahradí tělem
  setCursor(body_x[0], body_y[0]);
  Serial.write('O');
  // posunou se souřadnice v poli
  for (int i = body_len - 1; i > 0; i--)
  {
    body_x[i] = body_x[i - 1];
    body_y[i] = body_y[i - 1];
  }
  body_x[0] = hx;
  body_y[0] = hy;
  // vykreslí se hlava na nové pozici
  setCursor(hx, hy);
  Serial.print("☺");
  return r;
}

void setup() {
  Serial.begin(SERIAL_SPEED);
  // vytiskne pokyny
  clearScreen();
  Serial.println(F(" ☺oooo S N A K E ! OOOO☺ "));
  Serial.println(F("=========================="));
  Serial.println(F("Ovládání:"));
  Serial.println(F("W - nahoru"));
  Serial.println(F("S - dolů"));
  Serial.println(F("A - doleva"));
  Serial.println(F("D - doprava"));
  Serial.println();
  Serial.println(F("♣ - potrava"));
  Serial.println(F("♦ - jedovatý list"));
  Serial.println(F("▓ - zeď"));
  Serial.println();
  Serial.println(F("Stiskni mezerník pro start hry!"));
  waitState = true;
}

void loop() {
  int seed = 0;
  while (waitState)
  {
    randomSeed(seed++); // náhoda se odvíjí od okamžiku, kdy spustíme hru
    waitState = Serial.read() != 32;
    if (!waitState)
    {
      restartGame();
    }
     
  }
  // posune hada + určí, do čeho narazil
  int id = moveSnake();
  if ( id == ID_WALL)
  {
    Serial.print("Hlavou zeď neprorazíš!");
    waitState = true;
  }
  if ( id == ID_POISON)
  {
    Serial.print("Otrávil ses!");
    waitState = true;
  }
  if ( id == ID_SNAKE)
  {
    Serial.print("Sám sebe nežeru!");
    waitState = true;
  }
  if ( id == ID_LEAF)
  {
    // vygeneruje nový lístek (ten starý se překryl hlavičkou)
    newLeaf();
    // vygeneruje i jed
    newPoison();
    // prodlouží tělo, ovšem na maximálně SNAKE_MAX
    if (++body_len > SNAKE_MAX)
      body_len = SNAKE_MAX;
    // nový (poslední) článek má stejnou pozici jako předposlední
    body_x[body_len - 1] = body_x[body_len - 2];
    body_y[body_len - 1] = body_y[body_len - 2];
    // navýšení skóre
    score++;
    // výpis nového skóre
    setCursor(1, HEIGHT+3);
    Serial.print("Skóre: ");
    Serial.print(score);
  }  
  // hra se zrychluje se zvyšujícím se skóre
  int d = 300-score*4;
  if(d < 100)
    d = 100;
  delay(d);
  
  // zpracování vstupu 
  int p = Serial.read();
  while (p  != -1)
  {
    // pohyb doleva?
    if ((p == 'A' || p == 'a') && moveDir != MOVE_RIGHT)
    {
      moveDir = MOVE_LEFT;
    }
    // pohyb doprava
    if ((p == 'D' || p == 'd') && moveDir != MOVE_LEFT)
    {
      moveDir = MOVE_RIGHT;
    }
    // pohyb dolů?
    if ((p == 'S' || p == 's') && moveDir != MOVE_UP)
    {
      moveDir = MOVE_DOWN;
    }
    // pohyb nahoru?
    if ((p == 'W' || p == 'w') && moveDir != MOVE_DOWN)
    {
      moveDir = MOVE_UP;
    }
    p = Serial.read();
  }
}
